use crate::{
    entity::{Entity, EntityDatum, EntityPart, EntityPartVisitor, EntityVisitor},
    query,
    schema::collect_from_database,
};
use crate::{DBError, DBResult};

use std::{
    collections::HashMap,
    sync::{Arc, Mutex, Weak},
};

// ----------------------------------------------------------------------
// sqlite layer types
// ----------------------------------------------------------------------

pub(crate) type DBConnection = std::sync::Arc<Connection>;

pub struct Connection {
    conn: &'static mut sqlite::ConnectionThreadSafe,
    statement_cache: Mutex<HashMap<u64, sqlite::Statement<'static>>>,
}

impl Connection {
    pub fn open<U: AsRef<str>>(uri: U) -> Result<DBConnection, DBError> {
        match sqlite::Connection::open_thread_safe_with_flags(
            uri.as_ref(),
            sqlite::OpenFlags::new()
                .with_create()
                .with_full_mutex()
                .with_read_write(),
        ) {
            Ok(conn) => {
                let conn = Box::leak(Box::new(conn));
                Ok(Arc::new(Self {
                    conn,
                    statement_cache: Default::default(),
                }))
            }
            Err(e) => Err(DBError::Sqlite(e)),
        }
    }

    pub(crate) fn execute_raw(&self, sql: &str) -> Result<(), DBError> {
        Ok(self.conn.execute(sql)?)
    }

    pub(crate) fn with_prepared<R>(
        &self,
        hash_key: u64,
        build_query: impl Fn() -> String,
        run_query: impl Fn(&mut sqlite::Statement<'static>) -> DBResult<R>,
    ) -> DBResult<R> {
        todo!()
    }
}

impl Drop for Connection {
    fn drop(&mut self) {
        // first we drop the statement cache
        self.statement_cache
            .lock()
            .expect("couldn't get statement cache lock")
            .clear();
        // then we drop the connection
        unsafe {
            drop(Box::from_raw(self.conn));
        }
    }
}
