use std::{fmt::Debug, hash::Hash};

mod datum;

/// Integral identifier for an entity.
pub trait EntityID: 'static + PartialEq + Hash + PartialOrd + Debug + Copy {
    type Entity: Entity<ID = Self>;

    /// Construct from a raw integer.
    ///
    /// Do not use this unless you are _very_ certain it is what you want.
    fn from_raw(id: usize) -> Self;
    fn into_raw(self) -> usize;
}

/// Represents a data field in an Entity.
pub trait EntityDatum: 'static {
    fn sql_type() -> &'static str;

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, index: usize);

    fn accept_entity_visitor(_: &mut impl EntityVisitor) {}
}

/// A fixed-length list of EntityDatums, usually a tuple.
pub trait EntityDatumList {
    fn accept(&self, visitor: &mut impl EntityDatumListVisitor);
}

/// A walker for a EntityDatumList instance.
pub trait EntityDatumListVisitor {
    fn visit<ED: EntityDatum>(&mut self, datum: &ED);
}

/// A single data field in an Entity, automatically derived via `#[derive(Entity)]`.
pub trait EntityPart: 'static {
    type Datum: EntityDatum;
    type Entity: Entity;

    fn part_name() -> &'static str;
}

/// Visitor for traversing all `EntityPart`s in an `Entity`.
pub trait EntityPartVisitor {
    fn visit<EP: EntityPart>(&mut self) {}
    fn visit_datum<EP: EntityPart>(&mut self, _ctx: &'static str, _datum: &EP::Datum) {}
}

/// A single database entity, aka an object type that gets its own table.
pub trait Entity: 'static {
    type Uniques: EntityDatumList;
    type ID: EntityID<Entity = Self>;

    fn entity_name() -> &'static str;
    fn accept_part_visitor(visitor: &mut impl EntityPartVisitor);
    fn accept_part_visitor_ref(&self, visitor: &mut impl EntityPartVisitor);
}

/// Visitor for traversing all `Entity`s in a container.
pub trait EntityVisitor {
    fn visit<E: Entity>(&mut self);
}
