use crate::{
    db::DBConnection,
    entity::{Entity, EntityDatum, EntityPart, EntityPartVisitor},
    schema::IDMap,
};
use crate::{DBError, DBResult};
use std::hash::{Hash, Hasher};

#[derive(Hash)]
enum QueryType {
    Insert,
}

fn query_hash<E: Entity>(ctx: &'static str, qtype: QueryType) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();

    qtype.hash(&mut hasher);
    std::any::TypeId::of::<E>().hash(&mut hasher);
    ctx.hash(&mut hasher);

    hasher.finish()
}

pub fn insert<E: Entity>(map: &IDMap<E>, value: &E) -> DBResult<()> {
    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::Insert),
        || {
            let table_name = format!("{}_{}", map.ctx(), E::entity_name());

            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a>(&'a mut String, &'a mut String);
            impl<'a> EntityPartVisitor for PartNameVisitor<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    if self.0.len() != 0 {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.1.push_str("?");
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut part_names, &mut placeholders));

            format!(
                "insert into `{}` ({}) values ({})",
                table_name, part_names, placeholders
            )
        },
        |prepared| {
            struct PartBinder<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> EntityPartVisitor for PartBinder<'a, 'b> {
                fn visit_datum<EP: EntityPart>(&mut self, _ctx: &'static str, datum: &EP::Datum) {
                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(prepared, 1));

            prepared.next()?;

            DBResult::Ok(())
        },
    )
}
