use crate::{
    db::{Connection, DBConnection},
    entity::{Entity, EntityDatum, EntityVisitor},
    query,
};
use crate::{DBError, DBResult};

mod entity;
pub(crate) mod meta;
mod tests;

#[derive(Debug)]
struct ColumnInfo {
    name: &'static str,
    ty: String,
    fkey: Option<String>,
    unique: bool,
}

#[derive(Debug)]
struct TableInfo {
    table_name: String,
    columns: Vec<ColumnInfo>,
    dependencies: Vec<String>,
}

impl TableInfo {
    fn new(name: String) -> Self {
        TableInfo {
            table_name: name,
            columns: vec![],
            dependencies: vec![],
        }
    }

    fn build_creation_query(&self) -> String {
        let columns = self.columns.iter().map(|col| {
            format!(
                ", `{}` {}{}",
                col.name,
                col.ty,
                if col.unique { "unique" } else { "" }
            )
        });
        let fkeys = self.columns.iter().filter_map(|col| {
            Some(format!(
                ", foreign key(`{}`) references {}",
                col.name,
                col.fkey.as_ref()?
            ))
        });

        format!(
            "create table {} (`id` integer primary key{}{});",
            self.table_name,
            columns.collect::<String>(),
            fkeys.collect::<String>()
        )
    }
}

pub struct DatabaseSchema {
    signature: u64,
    queries: Vec<String>,
}

impl DatabaseSchema {
    const SCHEMA_SIGNATURE_KEY: &'static str = "schema_signature";
    pub fn check(&self, db: DBConnection) -> bool {
        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::build(db.clone());

        // check to see if the signature exists and matches
        metadb
            .kv_metastore
            .lookup_unique(&Self::SCHEMA_SIGNATURE_KEY.to_string())
            .ok()
            .flatten()
            .map(|kv| kv.value.parse::<u64>().unwrap_or(0) == self.signature)
            .unwrap_or(false)
    }

    pub fn create(&self, db: DBConnection) {
        for query in self.queries.iter() {
            db.execute_raw(query).expect("Couldn't run creation query!");
        }

        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::build(db.clone());

        for query in collect_from_database::<meta::MetadataDB>().queries.iter() {
            db.execute_raw(query)
                .expect("Couldn't run MetadataDB creation query!");
        }

        // store signature
        metadb
            .kv_metastore
            .insert(&meta::KV {
                key: Self::SCHEMA_SIGNATURE_KEY.into(),
                value: format!("{}", self.signature),
            })
            .expect("couldn't set schema signature");
    }
}

pub(crate) fn collect_from_database<DB: Database>() -> DatabaseSchema {
    struct IV(entity::EntityStateContainer);

    impl DatabaseItemVisitor for IV {
        fn visit<DI: DatabaseItem>(&mut self)
        where
            Self: Sized,
        {
            DI::accept_entity_visitor(&mut self.0.make_context(DI::item_key()));
        }
    }

    let mut iv = IV(entity::EntityStateContainer::default());

    DB::accept_item_visitor(&mut iv);

    // now to turn all that into a set of tables
    let mut tables = std::collections::HashMap::new();

    for state in iv.0.iter_states() {
        let table_name = format!("{}_{}", state.context, state.name);
        // we may end up visiting duplicate entities; skip them if so
        if tables.contains_key(&table_name) {
            continue;
        }
        let mut table = TableInfo::new(table_name.clone());
        for part in state.parts.iter() {
            match part.ty {
                entity::PartType::Datum(dtype) => {
                    table.columns.push(ColumnInfo {
                        name: part.name,
                        ty: dtype.into(),
                        fkey: None,
                        unique: false, // XXX
                    })
                }
                entity::PartType::Assoc(assoc_name) => {
                    let assoc_table_name = format!(
                        "{}_{}_assoc_{}_{}",
                        state.context, state.name, part.name, assoc_name
                    );
                    let mut assoc_table = TableInfo::new(assoc_table_name.clone());
                    assoc_table.dependencies.push(table_name.clone());
                    assoc_table
                        .dependencies
                        .push(format!("{}_{}", state.context, assoc_name));

                    assoc_table.columns.push(ColumnInfo {
                        name: "base".into(),
                        ty: "int".into(),
                        fkey: Some(format!("{}(`id`)", table_name)),
                        unique: false,
                    });

                    assoc_table.columns.push(ColumnInfo {
                        name: "target".into(),
                        ty: "int".into(),
                        fkey: Some(format!("{}_{}(`id`)", state.context, assoc_name)),
                        unique: false,
                    });

                    tables.insert(assoc_table_name, assoc_table);
                }
            }
        }

        tables.insert(table_name, table);
    }

    println!("collected base info from database: {:?}", iv.0);
    println!("derived table info: {:?}", tables);

    let mut tsort: topological_sort::TopologicalSort<&str> =
        topological_sort::TopologicalSort::new();
    for table in tables.values() {
        tsort.insert(table.table_name.as_str());

        for dep in table.dependencies.iter() {
            tsort.add_dependency(dep.as_str(), table.table_name.as_str());
        }
    }

    // this must be a stable hash function, so we very explicitly want to use a SipHasher with
    // known parameters
    #[allow(deprecated)]
    let mut signature_hasher = std::hash::SipHasher::new();
    use std::hash::{Hash, Hasher};

    let mut queries = vec![];

    loop {
        let mut table_list = tsort.pop_all();
        if table_list.len() == 0 {
            break;
        }
        // bring into stable ordering
        table_list.sort();

        for table_name in table_list.into_iter() {
            let table = tables.get(table_name).unwrap();
            let create_sql = table.build_creation_query();

            table_name.hash(&mut signature_hasher);
            create_sql.hash(&mut signature_hasher);
            queries.push(create_sql);
        }
    }

    // TODO: generate index schemas here

    DatabaseSchema {
        signature: signature_hasher.finish(),
        queries,
    }
}

// ----------------------------------------------------------------------
// Specification types
// ----------------------------------------------------------------------

pub struct IDMap<T: Entity> {
    conn: DBConnection,
    ctx: &'static str,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> IDMap<T> {
    pub fn build(db: DBConnection, ctx: &'static str) -> Self {
        Self {
            conn: db,
            ctx,
            _ghost: std::marker::PhantomData,
        }
    }

    pub(crate) fn conn(&self) -> &DBConnection {
        &self.conn
    }

    pub(crate) fn ctx(&self) -> &'static str {
        self.ctx
    }

    pub fn lookup_unique(&self, uniques: &T::Uniques) -> DBResult<Option<T>> {
        // TODO
        Ok(None)
    }

    pub fn insert(&self, value: &T) -> DBResult<()> {
        query::insert(self, value)
    }
}

pub struct Index<T: Entity, Key: EntityDatum> {
    _conn: DBConnection,
    _ghost: std::marker::PhantomData<(T, Key)>,
}

pub struct AssocSet<T: Entity> {
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> EntityDatum for AssocSet<T> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<T>();
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, index: usize) {
        todo!()
    }
}

pub trait DatabaseItem {
    fn item_key() -> &'static str;
    fn dependency_keys() -> &'static [&'static str];

    fn is_index() -> bool {
        false
    }
    fn index_over() -> &'static str {
        unreachable!()
    }
    fn index_columns() -> &'static [&'static str] {
        unreachable!()
    }

    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

pub trait DatabaseItemVisitor {
    fn visit<DI: DatabaseItem>(&mut self)
    where
        Self: Sized;
}

pub trait DatabaseSpec {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

impl<T: Entity> DatabaseSpec for IDMap<T> {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
        visitor.visit::<T>()
    }
}

pub trait Database {
    fn open_path<U: AsRef<str>>(uri: U) -> Result<Self, DBError>
    where
        Self: Sized,
    {
        Connection::open(uri).map(Self::build)
    }

    #[doc(hidden)]
    fn build(conn: DBConnection) -> Self
    where
        Self: Sized;

    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}
