mod manual_test_db {
    // simple hand-built database example
    #![allow(unused)]

    use crate::db::DBConnection;
    use crate::entity::{Entity, EntityID, EntityPart, EntityPartVisitor, EntityVisitor};
    use crate::schema::{Database, DatabaseItem, DatabaseItemVisitor, IDMap};

    struct SimpleEntity {
        name: String,
    }

    #[derive(Clone, Copy, PartialEq, PartialOrd, Debug, Hash)]
    struct SimpleEntityID(usize);

    impl EntityID for SimpleEntityID {
        type Entity = SimpleEntity;

        fn from_raw(id: usize) -> Self {
            Self(id)
        }
        fn into_raw(self) -> usize {
            self.0
        }
    }

    struct SimpleEntityName;
    impl EntityPart for SimpleEntityName {
        type Datum = String;
        type Entity = SimpleEntity;
        fn part_name() -> &'static str {
            "name"
        }
    }

    impl Entity for SimpleEntity {
        type Uniques = String;
        type ID = SimpleEntityID;

        fn entity_name() -> &'static str {
            "simple_entity"
        }
        fn accept_part_visitor(visitor: &mut impl EntityPartVisitor) {
            visitor.visit::<SimpleEntityName>();
        }
        fn accept_part_visitor_ref(&self, visitor: &mut impl EntityPartVisitor) {
            visitor.visit_datum::<SimpleEntityName>("name", &self.name);
        }
    }

    struct SimpleDatabase {
        strings: IDMap<SimpleEntity>,
    }

    impl Database for SimpleDatabase {
        fn build(conn: DBConnection) -> Self
        where
            Self: Sized,
        {
            Self {
                strings: IDMap::build(conn, "strings"),
            }
        }

        fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
        where
            Self: Sized,
        {
            struct SimpleDatabaseStringsItem;
            impl DatabaseItem for SimpleDatabaseStringsItem {
                fn item_key() -> &'static str {
                    "strings"
                }
                fn dependency_keys() -> &'static [&'static str] {
                    &[]
                }

                fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
                    visitor.visit::<SimpleEntity>();
                }
            }

            visitor.visit::<SimpleDatabaseStringsItem>();
        }
    }

    #[test]
    fn part_visitor() {
        struct V {
            v: Vec<std::any::TypeId>,
        }
        impl EntityPartVisitor for V {
            fn visit<EP: EntityPart>(&mut self) {
                self.v.push(std::any::TypeId::of::<EP>());
            }
        }

        let mut vis = V { v: vec![] };
        SimpleEntity::accept_part_visitor(&mut vis);
        assert_eq!(
            vis.v.as_slice(),
            &[std::any::TypeId::of::<SimpleEntityName>()]
        );
    }
}

mod derive_tests {
    #![allow(unused)]

    use crate::schema::{AssocSet, Database, IDMap};
    use microrm_macros::{Database, Entity};

    #[derive(Entity)]
    struct Role {
        title: String,
        permissions: String,
    }

    #[derive(Entity)]
    struct Person {
        #[unique]
        name: String,
        roles: AssocSet<Role>,
    }

    #[derive(Database)]
    struct PeopleDB {
        people: IDMap<Person>,
    }

    #[test]
    fn collect_test() {
        microrm::schema::collect_from_database::<PeopleDB>();
    }

    #[test]
    fn open_test() {
        // PeopleDB::open_path(std::path::Path::new(":memory:"));
        PeopleDB::open_path(std::path::Path::new("/tmp/schema.db").to_str().unwrap());
    }
}

/*
// pub trait

pub struct Skill { }
impl Entity for Skill {}

pub struct MenuItem {
    name: String,
    required_skills: Set<Skill>,
    cost: f32,
}

impl Entity for MenuItem {}

struct EmployeeShiftMap;

pub struct Employee {
    assigned_shifts: NamedMap<EmployeeShiftMap, Shift>,
}

impl Entity for Employee {}

pub struct Shift {
    name: String,
    employees: Set<Employee>,
}

impl Entity for Shift {}

pub struct FoodTruckSchema {
    menu: Collection<MenuItem>,
}
*/
